# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).


from plumbum import cli, local
from plumbum.commands.modifiers import FG
from pwd import getpwnam
import logging
import os

from ..common.api import logger, raise_error
from ..common.config import DockyConfig
from ..common.project import Project
from ..common.proxy import Proxy


class Docky(cli.Application):
    PROGNAME = "docky"
    VERSION = '5.0.1'
    SUBCOMMAND_HELPMSG = None

    force_env = cli.SwitchAttr(
        ["e", "env"],
        help="Environment flag",
        group = "Switches")

    def _run(self, cmd, retcode=FG):
        """Run a command in a new process and log it"""
        logger.debug(str(cmd).replace('/usr/local/bin/', ''))
        return cmd & retcode

    def _exec(self, cmd, args=[]):
        """Run a command in the same process and log it
        this will replace the current process by the cmd"""
        logger.debug(cmd + ' '.join(args))
        os.execvpe(cmd, [cmd] + args, local.env)

    def __init__(self, executable):
        super(Docky, self).__init__(executable)
        self.config = DockyConfig()
        if self.config.verbose:
            self.set_log_level()
            logger.debug(
                'Start in verbose mode. You can change the default '
                'value in ~/.docky/config.yml')

    @cli.switch("--verbose", help="Verbose mode", group = "Meta-switches")
    def set_log_level(self):
        logger.setLevel(logging.DEBUG)
        logger.debug('Verbose mode activated')


class DockySub(cli.Application):
    _project_specific = True

    def _exec(self, *args, **kwargs):
        self.parent._exec(*args, **kwargs)

    def _run(self, *args, **kwargs):
        self.parent._run(*args, **kwargs)

    def _init_project(self):
        self.project = Project(self.env, self.parent.config)
        self.project.build_network()
        self.compose = local['docker-compose'][
            '-f', self.project.compose_file_path,
            '--project-name', self.project.name]

    def main(self, *args, **kwargs):
        local.env['UID'] = str(getpwnam(local.env.user).pw_uid)
        self.env = self.parent.force_env or self.parent.config.env
        self.proxy = Proxy(self.parent.config)
        if self._project_specific:
            self._init_project()
        self._main(*args, **kwargs)


class DockySubNoProject(DockySub):
    _project_specific = False


DockySub.unbind_switches("--help-all", "-v", "--version")



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from plumbum import cli
from plumbum.cmd import git
from .base import Docky, DockySub, raise_error


@Docky.subcommand("migrate")
class DockyMigrate(DockySub):
    """TODO Migrate your odoo project"""
#
#    First you need to checkout the docky-upgrade template
#    available here : https://github.com/akretion/docky-upgrade
#    (It's a template a docky but based on open-upgrade'
#
#    Then go inside the repository clonned and launch the migration
#
#    * For migrating from 6.1 to 8.0 run:
#        docky migrate -b 7.0,8.0
#    * For migrating from 6.1 to 9.0 run:
#        docky migrate -b 7.0,8.0,9.0
#    * For migrating and loading a database run:
#        docky migrate -b 7.0,8.0 --db-file=tomigrate.dump
#
#    """

    db_file = cli.SwitchAttr(["db-file"])
    apply_branch = cli.SwitchAttr(
        ["b", "branch"],
        help="Branch to apply split by comma ex: 7.0,8.0",
        mandatory=True)
    _logs = []

    def log(self, message):
        print(message)
        self._logs.append(message)

    def _run_ak(self, *params):
        start = datetime.now()
        cmd = "ak " + " ".join(params)
        self.log("Launch %s" % cmd)
        self.compose("run", "odoo", "ak", *params)
        end = datetime.now()
        self.log("Run %s in %s" % (cmd, end-start))

    def _main(self):
        if self.main_service != 'odoo':
            raise_error("This command is used only for migrating odoo project")
        versions = self.apply_branch.split(',')
        logs = ["\n\nMigration Log Summary:\n"]
        first = True
        for version in versions:
            start = datetime.now()
            self._run(git["checkout", version])
            self._run_ak("build")
            if self.db_file and first:
                self._run_ak("db", "load", "--force", self.db_file)
                first = False
            self._run_ak("upgrade")
            self._run_ak("db", "dump", "--force", "migrated_%s.dump" % version)
            end = datetime.now()
            self.log("Migrate to version %s in %s" % (version, end-start))
        for log in self._logs:
            logger.info(log)



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from .base import Docky, DockySubNoProject, raise_error, logger
from plumbum import cli


@Docky.subcommand("proxy")
class DockyProxy(cli.Application):
    """Start/Restart/Stop your docky proxy"""

    def main(self, *args, **kwargs):
        self.force_env = self.parent.force_env
        self.env = self.parent.config.env
        self.config = self.parent.config
        if not self.nested_command:
            raise_error("Please specify an action, start/stop/restart/kill/ps")

@DockyProxy.subcommand("start")
class DockyProxyStart(DockySubNoProject):
    """Start your docky proxy"""

    def _main(self, *args):
        self.proxy.start()


@DockyProxy.subcommand("stop")
class DockyProxyStop(DockySubNoProject):
    """Stop your docky proxy"""

    def _main(self, *args):
        self.proxy.stop()


@DockyProxy.subcommand("restart")
class DockyProxyRestart(DockySubNoProject):
    """Restart your docky proxy"""

    def _main(self, *args):
        self.proxy.restart()


@DockyProxy.subcommand("ps")
class DockyProxyPs(DockySubNoProject):
    """Get the status of your docky proxy"""

    def _main(self, *args):
        self.proxy.status()


@DockyProxy.subcommand("kill")
class DockyProxyKill(DockySubNoProject):
    """Kill your docky proxy"""

    def _main(self, *args):
        self.proxy.kill()



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import os
import yaml
from .generator import GenerateComposeFile
from .api import logger


DEFAULT_CONF = {
    "verbose": True,
    "env": "dev",
    "network": {
        "name": "dy",
        "subnet": "172.30.0.0/16",
        "options": {
            "com.docker.network.bridge.name": "dy",
            "com.docker.network.bridge.host_binding_ipv4": "127.0.0.1",
        },
    },
    "proxy": {
        "custom_image": None,
        "name": "proxy-docky",
        "autostart": True,
    }
}


# TODO refactor code using local object of plumbum instead of os
class DockyConfig(object):

    def _get_home(self):
        return os.path.expanduser("~")

    def __init__(self):
        super(DockyConfig, self).__init__()
        self.home = self._get_home()
        docky_folder = os.path.join(self.home, '.docky')
        if not os.path.exists(docky_folder):
            os.makedirs(docky_folder)
        self.config_path = os.path.join(self.home, '.docky', 'config.yml')
        current_config = self._get_config()
        config = self._get_config_with_default_value(current_config)

        if current_config != config:
            self._update_config_file(config)
        self.env = config['env']
        self.verbose = config['verbose']
        self.network = config['network']
        self.proxy = config['proxy']

    def _get_config(self):
        if os.path.isfile(self.config_path):
            return yaml.safe_load(open(self.config_path, 'r'))
        else:
            return {}

    def _get_config_with_default_value(self, current_config):
        config = DEFAULT_CONF.copy()
        for key, value in DEFAULT_CONF.items():
            if key in current_config:
                config[key] = current_config[key]
        return config

    def _update_config_file(self, config):
        logger.info("The Docky Configuration have been updated, "
                    "please take a look to the new config file")
        config_file = open(self.config_path, 'w')
        config_file.write(yaml.dump(config, default_flow_style=False))
        logger.info("Update default config file at %s", self.config_path)



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from .base import Docky, DockySub
from compose.parallel import parallel_kill


@Docky.subcommand("kill")
class DockyKill(DockySub):
    """Kill all running container of the project"""

    def _main(self, *args):
        # docker compose do not kill the container odoo as is was run
        # manually, so we implement our own kill
        containers = self.project.get_containers()
        parallel_kill(containers, {'signal': 'SIGKILL'})



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from compose.cli.command import get_project
from compose.project import OneOffFilter
from plumbum import local
from plumbum.cli.terminal import ask
import docker
import yaml
import os

from .api import logger, raise_error
from .generator import GenerateComposeFile

client = docker.from_env()

DOCKY_PATH = 'docky.yml'
DOCKER_COMPOSE_PATH = 'docker-compose.yml'


class Project(object):

    def __init__(self, env, docky_config):
        self.env = env
        self.docky_config = docky_config
        self._parse_docky_file()
        self.compose_file_path = self._get_config_path()
        if self.env == 'dev':
            if not local.path(self.compose_file_path).isfile():
                self._generate_dev_docker_compose_file()
        self.name = self._get_project_name()
        self.loaded_config = None

    def _parse_docky_file(self):
        if os.path.isfile(DOCKY_PATH):
            config = yaml.safe_load(open(DOCKY_PATH, 'r'))
            self.service = config.get('service')
            self.user = config.get('user')
        else:
            raise_error(
                '%s file is missing. Minimal file is a yaml file with:\n'
                ' service: your_service\nex:\n service: odoo' % DOCKY_PATH)

    def _get_config_path(self):
        config_path = '.'.join([self.env, DOCKER_COMPOSE_PATH])
        if self.env == 'dev':
            return config_path
        elif local.path(config_path).is_file():
            return config_path
        elif local.path(DOCKER_COMPOSE_PATH).is_file():
            return DOCKER_COMPOSE_PATH
        else:
            raise_error(
                "There is not %s.%s or %s file, please add one"
                % (self.env, DOCKER_COMPOSE_PATH, DOCKER_COMPOSE_PATH))

    def _generate_dev_docker_compose_file(self):
        generate = ask(
            "There is not dev.docker-compose.yml file.\n"
            "Do you want to generate one automatically",
            default=True)
        if generate:
            GenerateComposeFile(self.service).generate()
        else:
            raise_error("No dev.docker-compose.yml file, abort!")

    def _get_project_name(self):
        return local.env.get(
            'COMPOSE_PROJECT_NAME',
            '%s_%s' % (local.env.user, local.cwd.name)
        )

    def get_containers(self, service=None):
        project = get_project(
            '.', [self.compose_file_path],
            project_name=self.name)
        kwargs = {'one_off': OneOffFilter.include}
        if service:
            kwargs['service_names'] = [service]
        return project.containers(**kwargs)

    @property
    def config(self):
        if not self.loaded_config:
            self.loaded_config = yaml.safe_load(
                open(self.compose_file_path, 'r'))
        return self.loaded_config

    def show_access_url(self):
        for name, service in self.config['services'].items():
            for env in service.get('environment', []):
                if 'VIRTUAL_HOST=' in env:
                    dns = env.replace('VIRTUAL_HOST=', '')
                    logger.info(
                        "The service %s is accessible on http://%s"
                        % (name, dns))

    def create_volume(self):
        for name, service in self.config['services'].items():
            if 'volumes' in service:
                for volume_path in service['volumes']:
                    volume = volume_path.split(':')[0]
                    if volume.startswith('.') or volume.startswith('/'):
                        path = local.path(volume)
                        if not path.exists():
                            logger.info(
                                "Create missing directory %s for service %s",
                                volume, name)
                            local.path(volume).mkdir()

    def build_network(self):
        network = self.docky_config.network
        if not network:
            logger.info("No network define, skip it")
        gateway = '.'.join(network['subnet'].split('.')[0:3] + ["1"])
        for net in client.networks.list(network['name']):
            if net.name == network['name']:
                return
        ipam_pool = docker.types.IPAMPool(
            subnet=network['subnet'],
            iprange=network['subnet'],
            gateway=gateway,
        )
        ipam_config = docker.types.IPAMConfig(
            pool_configs=[ipam_pool])

        logger.info("Create '.%s' network" % network['name'])

        client.networks.create(
            network['name'],
            driver="bridge",
            ipam=ipam_config,
            options=network['options'],
        )



#!/usr/bin/env python
# coding: utf-8

import yaml
import pkg_resources
from plumbum.cli.terminal import ask
from plumbum import local
from slugify import slugify


class IndentDumper(yaml.Dumper):

    def increase_indent(self, flow=False, indentless=False):
        return super(IndentDumper, self).increase_indent(flow, False)


class GenerateComposeFile(object):

    def __init__(self, service):
        super(GenerateComposeFile, self).__init__()
        # Do not use os.path.join()
        self.service = service
        resource_path = '../template/%s.docker-compose.yml' % service
        template = pkg_resources.resource_stream(__name__, resource_path)
        self.config = yaml.safe_load(template)

    def _ask_optional_service(self):
        """Container can be set as optional by adding the key
        "optional" and "description". This method will ask the user to
        use or not this optional container"""
        answer = {}
        for name, config in self.config['services'].copy().items():
            if config.get('optional'):
                option = config['optional']
                if not option in answer:
                    answer[option] = ask(
                        "%s. Do you want to install it"
                        % option, default=False)
                if answer[option]:
                    # remove useless docker compose key
                    del self.config['services'][name]['optional']
                    if not 'links' in self.config['services'][self.service]:
                        self.config['services'][self.service]['links'] = []
                    self.config['services'][self.service]['links'].append(name)
                else:
                    del self.config['services'][name]

    def _add_container_name(self):
        project_name = slugify(local.cwd.name)
        for name, config in self.config['services'].items():
            expose = config.pop('expose', False)
            if expose:
                if name != self.service:
                    dns = "%s.%s.dy" %(name, project_name)
                else:
                    dns = "%s.dy" % project_name
                if not 'environment' in config:
                    config['environment'] = []
                config['environment'].append("VIRTUAL_HOST=%s" % dns)
                config['environment'].append("VIRTUAL_PORT=%s" % expose)
                config['networks'] = {}
                config['networks']['default'] = {}
                config['networks']['default']['aliases'] = []
                config['networks']['default']['aliases'].append(dns)

    def _update_config_file(self):
        self._ask_optional_service()
        self._add_container_name()

    def generate(self):
        self._update_config_file()
        with open('dev.docker-compose.yml', 'w') as dc_tmp_file:
            dc_tmp_file.write(yaml.dump(
                self.config, Dumper=IndentDumper, default_flow_style=False))



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).


from plumbum import cli
from .base import Docky, DockySub, raise_error


class DockyExec(object):

    root = cli.Flag(
        ["root"],
        help="Run or open as root",
        group = "Meta-switches")

    def _use_specific_user(self):
        return not self.root and self.project.user


@Docky.subcommand("run")
class DockyRun(DockySub, DockyExec):
    """Start services and enter in your dev container"""

    def _check_running(self):
        if self.project.get_containers(service=self.project.service):
            raise_error("This container is already running, kill it or "
                        "use open to go inside")

    def _main(self, *optionnal_command_line):
        self._check_running()
        self.proxy.start_if_needed()
        if self._use_specific_user():
            cmd = ['gosu', self.project.user]
        else:
            cmd = []
        if not optionnal_command_line:
            cmd.append('bash')
        else:
            cmd += list(optionnal_command_line)
        # Remove useless dead container before running a new one
        self._run(self.compose['rm', '-f'])
        self.project.show_access_url()
        self.project.create_volume()
        self._exec('docker-compose', [
            '-f', self.project.compose_file_path,
            '--project-name', self.project.name,
            'run', '--rm', '--service-ports', '--use-aliases',
            self.project.service] + cmd)


@Docky.subcommand("open")
class DockyOpen(DockySub, DockyExec):
    """Open a new session inside your dev container"""

    # Patch compose service to be make it working with docker-compose run

    def _main(self, service=None):
        cmd = ['bash']
        if not service:
            service = self.project.service
            if self._use_specific_user():
                cmd = ['gosu', self.project.user, 'bash']
        self._exec('dcpatched', [
            '-f', self.project.compose_file_path,
            '--project-name', self.project.name,
            'exec', service] + cmd)



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

from .base import Docky, DockySub


class DockyForward(DockySub):
    _cmd = None

    def _main(self, *args):
        cmd = self._cmd.split(' ')
        if args:
            cmd.append(*args)
        return self._run(self.compose[cmd])


@Docky.subcommand("build")
class DockyBuild(DockyForward):
    """Build or rebuild services"""
    _cmd = "build"


@Docky.subcommand("up")
class DockyUp(DockyForward):
    """Start all services in detached mode"""
    _cmd = "up -d"

    def _main(self, *args):
        self.project.show_access_url()
        self.project.create_volume()
        self.proxy.start_if_needed()
        return super(DockyUp, self)._main(*args)


@Docky.subcommand("down")
class DockyDown(DockyForward):
    """Stop all services"""
    _cmd = "down"


@Docky.subcommand("ps")
class DockyPs(DockyForward):
    """List containers"""
    _cmd = "ps"


@Docky.subcommand("logs")
class DockyLogs(DockyForward):
    """View output from containers"""
    _cmd = "logs -f --tail=100"


@Docky.subcommand("pull")
class DockyPull(DockyForward):
    """Pulls service images"""
    _cmd = "pull"


@Docky.subcommand("restart")
class DockyRestart(DockyForward):
    """Restart service"""
    _cmd = "restart"





# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).


import logging
import sys

logger = logging.getLogger('docky')
formatter = logging.Formatter("%(message)s")
logger.setLevel(logging.INFO)

# Optionnal code for colorized log
from rainbow_logging_handler import RainbowLoggingHandler
handler = RainbowLoggingHandler(
    sys.stderr,
    color_message_info = ('green' , None , True),
)
handler.setFormatter(formatter)
logger.addHandler(handler)


def raise_error(message):
    logger.error(message)
    sys.exit(0)



from setuptools import setup, find_packages
import codecs
import os
import re

def read(*parts):
    path = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(path, encoding='utf-8') as fobj:
        return fobj.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"VERSION = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

setup(
    name='docky',
    version=find_version("docky", "cmd", "base.py"),
    author='Akretion',
    author_email='contact@akretion.com',
    url='https://github.com/akretion/docky/',
    description='Make developpement with docker simply',
    license="AGPLv3+",
    long_description=open('README.rst').read(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        r.strip() for r in open('requirements.txt').read().splitlines() ],
    entry_points="""
    [console_scripts]
    docky=docky.main:main
    dcpatched=docky.dcpatched:main
    """,
    include_package_data=True,
    packages = find_packages() + ['docky'],
    zip_safe=False,
)



# coding: utf-8
from compose import cli
from compose.service import Service
from compose.cli.main import TopLevelCommand
from compose.const import LABEL_CONTAINER_NUMBER


# This is a patched version of docker-compose
# We patch it in order to be able to
# - do a docker-compose exec on a running container (launch with run)


get_ori_container = Service.get_container


def get_run_container(self, number=1):
    # search for container running in background
    for container in self.containers(
            labels=['{0}={1}'.format(LABEL_CONTAINER_NUMBER, number)]):
        return container

    # search for container running with "run" cmd
    for container in self.containers(
            one_off=True,
            labels=['{0}={1}'.format(LABEL_CONTAINER_NUMBER, number)]):
        return container
    raise ValueError("No container found for %s_%s" % (self.name, number))


def main():
    Service.get_container = get_run_container
    cli.main.main()



# -*- coding: utf-8 -*-
# Copyright 2018 Akretion (http://www.akretion.com).
# @author Sébastien BEAU <sebastien.beau@akretion.com>
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import docker
from .api import logger

client = docker.from_env()

DOCKY_PROXY_DEFAULT_IMAGE = "quay.io/akretion/docky-proxy:20180723"


class Proxy(object):

    def __init__(self, docky_config):
        super(Proxy, self).__init__()
        self.network = docky_config.network
        config = docky_config.proxy
        self.image = config.get('custom_image') or DOCKY_PROXY_DEFAULT_IMAGE
        self.name = config['name']
        self.autostart = config['autostart']
        self.container = self._get_container()

    def _get_container(self):
        container = client.containers.list(
            all=True,
            filters={'name': self.name})
        return container[0] if container else None

    def _restart(self):
        logger.info("Restart docky proxy")
        self.container.restart()

    def _start(self):
        logger.info("Start Docky proxy")
        client.containers.run(
            self.image,
            hostname=self.name,
            name=self.name,
            network_mode=self.network['name'],
            volumes=[
                "/var/run/docker.sock:/tmp/docker.sock:ro",
                ],
                detach=True)

    def start_if_needed(self):
        if self.autostart and not self._is_running():
            self.start()

    def _is_running(self):
        return self.container and self.container.status == 'running'

    def start(self):
        if not self.container:
            self._start()
        elif self.container.status != 'running':
            logger.info(
                "A container already exist but is not running. Restart it")
            self._restart()
        else:
            logger.info("Proxy is already running, skip")

    def stop(self):
        if not self._is_running():
            logger.info("Proxy is already stopped, skip")
        else:
            logger.info("Stop the Proxy")
            self.container.stop()

    def restart(self):
        if not self.container:
            logger.info("There is no proxy start it")
            self._start()
        else:
            self._restart()

    def status(self):
        if self.container:
            status = self.container.status
        else:
            status = 'container do not exist'
        logger.info('Proxy status: %s', status)

    def kill(self):
        if not self.container:
            logger.info('Container do not exist, skip')
        else:
            if self.container.status != 'exited':
                logger.info('Kill docky proxy')
                self.container.kill()
            logger.info('Remove docky container proxy')
            self.container.remove()
